<?php

namespace Drupal\clockify;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DateFormatter;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use Symfony\Component\DependencyInjection\ContainerInterface;

define('DATETIME_DATETIME_STORAGE_FORMAT', 'Y-m-d\\TH:i:s');

/**
 * Clockify Controller to fetch data.
 */
class ClockifyController {

  /**
   * Api key.
   *
   * @var string
   */
  private $globalApi;

  /**
   * The api url.
   *
   * @var string
   */
  private $apiUrl = 'https://api.clockify.me/api/v1/';

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The HTTP client.
   *
   * @var GuzzleHttp\Client
   */
  protected $client;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The date formatter.
   *
   * @var \Drupal\Core\Datetime\DateFormatter
   */
  protected $dateFormatter;

  /**
   * The form.
   *
   * @var string
   */
  protected $loadFrom;

  /**
   * ClockifyController constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \GuzzleHttp\Client $httpClient
   *   The HTTP client service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Datetime\DateFormatter $dateFormatter
   *   The date formatter.
   * @param string $request_from
   *   The request form.
   */
  public function __construct(ConfigFactoryInterface $configFactory, Client $httpClient, EntityTypeManagerInterface $entityTypeManager, DateFormatter $dateFormatter, $request_from = FALSE) {
    $this->configFactory = $configFactory->get('clockify.settings');
    $this->globalApi = $this->configFactory->get('api_key');
    $this->client = $httpClient;
    $this->loadFrom = $request_from;
    $this->entityTypeManager = $entityTypeManager;
    $this->dateFormatter = $dateFormatter;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, $request_from = FALSE) {
    return new static(
      $container->get('config.factory'),
      $container->get('http_client'),
      $container->get('entity_type.manager'),
      $container->get('date.formatter'),
      $request_from
    );
  }

  /**
   * Loads workspaces accessible for the configured user.
   *
   * @return array
   *   Array containing the workspaces and its members.
   */
  public function getWorkspaces() {
    return $this->apiCall('workspaces/');
  }

  /**
   * Gets members form api for given workspace.
   *
   * @param string $workspace_id
   *   The workspace id used on clockify.
   *
   * @return array
   *   Contains the members and some basic information from them.
   */
  public function getMembers($workspace_id) {
    return $this->apiCall('workspaces/' . $workspace_id . '/users');
  }

  /**
   * Gets tags for given workspace.
   *
   * @param string $workspace_id
   *   The workspace id used on Clockify.
   */
  public function getTags($workspace_id) {
    $tags = $this->apiCall('/workspaces/' . $workspace_id . '/tags');
    $taxonomyTermStorage = $this->entityTypeManager->getStorage('taxonomy_term');
    $clockifyStorage = $this->entityTypeManager->getStorage('clockify');
    foreach ($tags as $value) {
      $query = $taxonomyTermStorage->getQuery();
      $query->accessCheck(FALSE);
      $query->condition('field_clockify_id', $value->id);
      $result = $query->execute();

      if (!$result) {
        $query = $clockifyStorage->getQuery();
        $query->accessCheck(FALSE);
        $query->condition('type', 'workspace');
        $query->condition('field_clockify_id', $value->workspaceId);
        $workspace = $query->execute();
        $workspace = reset($workspace);
        $taxonomyTermStorage->create([
          'name' => $value->name,
          'vid' => 'clockify_tags',
          'field_workspace' => [['target_id' => $workspace]],
          'field_clockify_id' => $value->id,
        ])->save();
      }

    }
  }

  /**
   * Loads the projects for any given workspace.
   *
   * @param string $workspace_id
   *   The workspace id used on clockify.
   *
   * @return array
   *   Contains information about the project.
   */
  public function getProjects($workspace_id) {
    return $this->apiCall('workspaces/' . $workspace_id . '/projects');
  }

  /**
   * Loads projects for the given workspaces.
   */
  public function saveWorkspaces() {
    $workspaces = $this->getWorkspaces();
    $clockifyStorage = $this->entityTypeManager->getStorage('clockify');
    foreach ($workspaces as $workspace) {
      $members = $this->saveMembers($workspace->id);
      $this->getTags($workspace->id);
      $query = $clockifyStorage->getQuery();
      $query->accessCheck(FALSE);
      $query->condition('type', 'workspace');
      $query->condition('field_clockify_id', $workspace->id);
      $result = $query->execute();
      if (!$result) {
        $values = [
          'type' => 'workspace',
          'name' => $workspace->name,
          'field_clockify_id' => $workspace->id,
          'field_members' => $members,
        ];
        $ws = $this->entityTypeManager->getStorage('clockify')->create($values);
        $ws->save();
        $query = $clockifyStorage->getQuery();
        $query->accessCheck(FALSE);
        $query->condition('field_clockify_id', $workspace->id);
        $result = $query->execute();
        $this->saveProjects($workspace->id, $ws->id());
      }
      else {
        $this->saveProjects($workspace->id, reset($result));
      }
      foreach ($workspace->memberships as $member) {
        $this->saveTimeEntry($workspace->id, $member->userId);
      }
    }
  }

  /**
   * Saves projects returnd by the api.
   *
   * @param string $workspace_id
   *   The clockify if of the workspace.
   * @param string $wid
   *   The drupal id of the workspace.
   */
  public function saveProjects($workspace_id, $wid) {
    $projects = $this->getProjects($workspace_id);
    $clockifyStorage = $this->entityTypeManager->getStorage('clockify');
    foreach ($projects as $project) {
      $query = $clockifyStorage->getQuery();
      $query->accessCheck(FALSE);
      $query->condition('type', 'project');
      $query->condition('field_clockify_id', $project->id);
      $result = $query->execute();
      if (!reset($result)) {
        $members = $this->loadMembers($project->memberships);
        $values = [
          'type' => 'project',
          'field_clockify_id' => $project->id,
          'name' => $project->name,
          'field_client_name' => $project->clientName,
          'field_color' => $project->color,
          'field_workspace' => $wid,
          'field_members' => $members,
        ];
        $pr = $this->entityTypeManager->getStorage('clockify')->create($values);
        $pr->save();
      }

    }
  }

  /**
   * Loads members for a given project.
   *
   * @param array $members
   *   Array containing the clockify user ids of members.
   *
   * @return array
   *   Array containing the user ids of members.
   */
  public function loadMembers(array $members) {
    $uids = [];
    $userStorage = $this->entityTypeManager->getStorage('user');
    foreach ($members as $member) {
      $query = $userStorage->getQuery();
      $query->accessCheck(FALSE);
      $query->condition('field_clockify_id', $member->userId);
      $result = $query->execute();
      if (count($result) > 0) {
        $uids[] = reset($result);
      }
    }
    return $uids;
  }

  /**
   * Loads and saves useres for the given workspace.
   *
   * @param string $workspace_id
   *   The ID of the workspace.
   *
   * @return array
   *   Array containing the user ids for the given workspace.
   */
  public function saveMembers($workspace_id) {
    $users = [];
    $members = $this->getMembers($workspace_id);
    $userStorage = $this->entityTypeManager->getStorage('user');
    foreach ($members as $member) {
      $query = $userStorage->getQuery();
      $query->accessCheck(FALSE);
      $query->condition('field_clockify_id', $member->id);
      $entity_ids = $query->execute();
      if (count($entity_ids) === 1) {
        $users[] = reset($entity_ids);
      }
      else {
        $query = $userStorage->getQuery();
        $query->accessCheck(FALSE);
        $query->condition('mail', $member->email);
        $entity_ids = $query->execute();
        if (count($entity_ids) === 1) {
          $uid = reset($entity_ids);
          $user = $userStorage->load($uid);
          $user->set('field_clockify_id', $member->id);
          $user->save();
          $users[] = $user->id();
        }
        else {
          $values = [
            'name' => $member->name,
            'mail' => $member->email,
            'field_clockify_id' => $member->id,
            'status' => 1,
          ];
          $account = $this->entityTypeManager->getStorage('user')->create($values);
          $account->save();
          $users[] = $account->id();
        }
      }
    }
    return $users;
  }

  /**
   * Gathers and saves time entryes.
   *
   * @param string $workspace_id
   *   The workspace id to gather from.
   * @param string $member_id
   *   The member id to gather from.
   */
  public function saveTimeEntry($workspace_id, $member_id) {
    $taxonomyTermStorage = $this->entityTypeManager->getStorage('taxonomy_term');
    $clockifyStorage = $this->entityTypeManager->getStorage('clockify');
    $userStorage = $this->entityTypeManager->getStorage('user');
    $path = 'workspaces/' . $workspace_id . '/user/' . $member_id . '/time-entries';
    $page = 1;
    $page_size = 20;
    if ($this->loadFrom) {
      $parameters = [
        'start' => $this->loadFrom,
        'page-size' => $page_size,
        'page' => $page,
      ];
      $query_parameters = UrlHelper::buildQuery($parameters);
      $path = $path . '?' . $query_parameters;
    }
    do {
      $time_tracking = $this->apiCall($path);
      foreach ($time_tracking as $time) {
        $query = $clockifyStorage->getQuery();
        $query->accessCheck(FALSE);
        $query->condition('type', 'time_entry');
        $query->condition('field_clockify_id', $time->id);
        $result = $query->execute();

        $start = strtotime($time->timeInterval->start);
        $start = $this->dateFormatter->format($start, 'custom', DATETIME_DATETIME_STORAGE_FORMAT);
        $end = strtotime($time->timeInterval->end);
        if (is_numeric($end)) {
          $end = $this->dateFormatter->format($end, 'custom', DATETIME_DATETIME_STORAGE_FORMAT);
        }
        else {
          $end = '';
        }
        $tags = [];
        if (!empty($time->tagIds)) {
          foreach ($time->tagIds as $tag) {
            $tag_query = $taxonomyTermStorage->getQuery();
            $tag_query->accessCheck(FALSE);
            $tag_query->condition('field_clockify_id', $tag);
            $tag_result = $tag_query->execute();
            if ($tag_result) {
              $tags[] = ['target_id' => reset($tag_result)];
            }
          }
        }
        if (!$result) {
          $time_owner = '';
          $user = $userStorage->getQuery();
          $user->accessCheck(FALSE);
          $user->condition('field_clockify_id', $time->userId);
          $user = $user->execute();
          if (reset($user)) {
            $time_owner = reset($user);
          }

          $values = [
            'type' => 'time_entry',
            'name' => substr($time->description, 0, 15),
            'field_description' => $time->description,
            'field_clockify_id' => $time->id,
            'field_start_time' => $start,
            'field_end_time' => $end,
            'field_workspace' => $this->getEntity('workspace', $time->workspaceId),
            'field_project' => $time->projectId ? $this->getEntity('project', $time->projectId) : '',
            'field_owner' => $time_owner,
            'field_duration' => $time->timeInterval->duration,
            'field_tags' => $tags ? $tags : [],
          ];
          $tr = $this->entityTypeManager->getStorage('clockify')->create($values);
          $tr->save();
        }
        else {
          $entry = $clockifyStorage->load(reset($result));
          $entry->set('field_start_time', $start);
          $entry->set('field_end_time', $end);
          $entry->set('field_duration', $time->timeInterval->duration);
          if ($tags) {
            $entry->set('field_tags', $tags);
          }
          $entry->save();
        }
      }
    } while (count($time_tracking) === $page_size);
  }

  /**
   * Load entity of type with id.
   *
   * @param string $type
   *   The type of the entity, ex: workspace, project or time_entry.
   * @param string $clockify_id
   *   The clockify id of the given content.
   *
   * @return sting
   *   The drupal entity id of the content.
   */
  public function getEntity(string $type, string $clockify_id) {
    $clockifyStorage = $this->entityTypeManager->getStorage('clockify');
    $query = $clockifyStorage->getQuery();
    $query->accessCheck(FALSE);
    $query->condition('type', $type);
    $query->condition('field_clockify_id', $clockify_id);
    $result = $query->execute();
    if (count($result) > 0) {
      return reset($result);
    }
    else {
      return '';
    }
  }

  /**
   * Loads members drupal ids.
   *
   * @param array $members
   *   Array containing the members returned by the clockify api.
   *
   * @return array
   *   Contains the drupal ids of members.
   */
  public function getMemberIds($members) {
    $ids = [];
    $userStorage = $this->entityTypeManager->getStorage('user');
    foreach ($members as $member) {
      $query = $userStorage->getQuery();
      $query->accessCheck(FALSE);
      $query->condition('field_clockify_id', $member->id);
      $uid = $query->execute();
      $ids[] = reset($uid);
    }
    return $ids;
  }

  /**
   * Calls the clockify api.
   *
   * @param string $url
   *   Contains the endpoint to call, ex /workspaces/{workspace_id}.
   * @param string $method
   *   Contains the method for the api call.
   *
   * @return array
   *   Returns the decoded response from the api call.
   */
  private function apiCall($url, $method = 'GET') {
    try {
      $path = $this->apiUrl . $url;
      $request = $this->client->request($method, $path, [
        'headers' => ['X-Api-Key' => $this->globalApi],
      ]);
      $file_contents = $request->getBody()->getContents();
      $result = json_decode($file_contents);
    }
    catch (ClientException $exception) {
      $result = [];
    }
    return $result;
  }

}
